package amd64

import (
	"fmt"
	"math"
	"testing"

	"github.com/tetratelabs/wazero/internal/asm"
	"github.com/tetratelabs/wazero/internal/testing/require"
)

func TestAssemblerImpl_Assemble_NOPPadding(t *testing.T) {
	t.Run("non relative jumps", func(t *testing.T) {
		tests := []struct {
			name     string
			setupFn  func(assembler Assembler)
			expected []byte
		}{
			{
				name: "RET",
				setupFn: func(assembler Assembler) {
					for i := 0; i < 128; i++ {
						assembler.CompileStandAlone(RET)
					}
				},
				expected: []byte{
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3,
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0x90,
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3,
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0x90, 0xc3, 0xc3,
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3,
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0x90, 0xc3, 0xc3, 0xc3, 0xc3,
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3,
					0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0x90, 0xc3, 0xc3, 0xc3, 0xc3,
				},
			},
			{
				name: "JMP to register",
				setupFn: func(assembler Assembler) {
					for i := 0; i < 128; i++ {
						assembler.CompileJumpToRegister(JMP, RegAX)
					}
				},
				expected: []byte{
					0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0x66, 0x90, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0, 0x66, 0x90, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0, 0xff, 0xe0, 0x66, 0x90, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0x66, 0x90, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0x66, 0x90, 0xff, 0xe0, 0xff, 0xe0, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0x66, 0x90, 0xff, 0xe0, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0x66, 0x90, 0xff,
					0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0x66,
					0x90, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0, 0xff, 0xe0,
					0xff, 0xe0,
				},
			},
			{
				name: "JMP to memory",
				setupFn: func(assembler Assembler) {
					for i := 0; i < 128; i++ {
						assembler.CompileJumpToMemory(JMP, RegAX, 10)
					}
				},
				expected: []byte{
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0x66, 0x90,
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa,
					0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff,
					0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0x66, 0x90, 0xff, 0x60, 0xa, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60, 0xa, 0xff, 0x60,
					0xa, 0xff, 0x60, 0xa,
				},
			},
			{
				name: "JMP to memory large offset",
				setupFn: func(assembler Assembler) {
					for i := 0; i < 128; i++ {
						assembler.CompileJumpToMemory(JMP, RegAX, math.MaxInt32)
					}
				},
				expected: []byte{
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0x66, 0x90,
					0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff, 0xff, 0x7f, 0xff, 0xa0, 0xff, 0xff,
					0xff, 0x7f,
				},
			},
		}

		for _, tt := range tests {
			tc := tt
			t.Run(tc.name, func(t *testing.T) {
				a := NewAssembler()
				tc.setupFn(a)

				actual, err := a.Assemble()
				require.NoError(t, err)

				require.Equal(t, tc.expected, actual)
			})
		}
	})

	t.Run("relative jumps", func(t *testing.T) {
		tests := []struct {
			jmpInst  asm.Instruction
			backward bool
			exp      []byte
		}{
			{
				backward: false, jmpInst: JMP, exp: []byte{
					0xc3, 0xe9, 0xe4, 0x1, 0x0, 0x0, 0xe9, 0xdf, 0x1, 0x0, 0x0, 0xe9, 0xda, 0x1, 0x0, 0x0, 0xe9, 0xd5,
					0x1, 0x0, 0x0, 0xe9, 0xd0, 0x1, 0x0, 0x0, 0xe9, 0xcb, 0x1, 0x0, 0x0, 0x90, 0xe9, 0xc5, 0x1, 0x0,
					0x0, 0xe9, 0xc0, 0x1, 0x0, 0x0, 0xe9, 0xbb, 0x1, 0x0, 0x0, 0xe9, 0xb6, 0x1, 0x0, 0x0, 0xe9, 0xb1,
					0x1, 0x0, 0x0, 0xe9, 0xac, 0x1, 0x0, 0x0, 0x66, 0x90, 0xe9, 0xa5, 0x1, 0x0, 0x0, 0xe9, 0xa0, 0x1,
					0x0, 0x0, 0xe9, 0x9b, 0x1, 0x0, 0x0, 0xe9, 0x96, 0x1, 0x0, 0x0, 0xe9, 0x91, 0x1, 0x0, 0x0, 0xe9,
					0x8c, 0x1, 0x0, 0x0, 0x66, 0x90, 0xe9, 0x85, 0x1, 0x0, 0x0, 0xe9, 0x80, 0x1, 0x0, 0x0, 0xe9, 0x7b,
					0x1, 0x0, 0x0, 0xe9, 0x76, 0x1, 0x0, 0x0, 0xe9, 0x71, 0x1, 0x0, 0x0, 0xe9, 0x6c, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xe9, 0x65, 0x1, 0x0, 0x0, 0xe9, 0x60, 0x1, 0x0, 0x0, 0xe9, 0x5b, 0x1, 0x0, 0x0, 0xe9,
					0x56, 0x1, 0x0, 0x0, 0xe9, 0x51, 0x1, 0x0, 0x0, 0xe9, 0x4c, 0x1, 0x0, 0x0, 0x66, 0x90, 0xe9, 0x45,
					0x1, 0x0, 0x0, 0xe9, 0x40, 0x1, 0x0, 0x0, 0xe9, 0x3b, 0x1, 0x0, 0x0, 0xe9, 0x36, 0x1, 0x0, 0x0, 0xe9,
					0x31, 0x1, 0x0, 0x0, 0xe9, 0x2c, 0x1, 0x0, 0x0, 0x66, 0x90, 0xe9, 0x25, 0x1, 0x0, 0x0, 0xe9, 0x20, 0x1,
					0x0, 0x0, 0xe9, 0x1b, 0x1, 0x0, 0x0, 0xe9, 0x16, 0x1, 0x0, 0x0, 0xe9, 0x11, 0x1, 0x0, 0x0, 0xe9, 0xc,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xe9, 0x5, 0x1, 0x0, 0x0, 0xe9, 0x0, 0x1, 0x0, 0x0, 0xe9, 0xfb, 0x0, 0x0,
					0x0, 0xe9, 0xf6, 0x0, 0x0, 0x0, 0xe9, 0xf1, 0x0, 0x0, 0x0, 0xe9, 0xec, 0x0, 0x0, 0x0, 0x66, 0x90, 0xe9,
					0xe5, 0x0, 0x0, 0x0, 0xe9, 0xe0, 0x0, 0x0, 0x0, 0xe9, 0xdb, 0x0, 0x0, 0x0, 0xe9, 0xd6, 0x0, 0x0, 0x0,
					0xe9, 0xd1, 0x0, 0x0, 0x0, 0xe9, 0xcc, 0x0, 0x0, 0x0, 0x66, 0x90, 0xe9, 0xc5, 0x0, 0x0, 0x0, 0xe9, 0xc0,
					0x0, 0x0, 0x0, 0xe9, 0xbb, 0x0, 0x0, 0x0, 0xe9, 0xb6, 0x0, 0x0, 0x0, 0xe9, 0xb1, 0x0, 0x0, 0x0, 0xe9,
					0xac, 0x0, 0x0, 0x0, 0x66, 0x90, 0xe9, 0xa5, 0x0, 0x0, 0x0, 0xe9, 0xa0, 0x0, 0x0, 0x0, 0xe9, 0x9b, 0x0,
					0x0, 0x0, 0xe9, 0x96, 0x0, 0x0, 0x0, 0xe9, 0x91, 0x0, 0x0, 0x0, 0xe9, 0x8c, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xe9, 0x85, 0x0, 0x0, 0x0, 0xe9, 0x80, 0x0, 0x0, 0x0, 0xeb, 0x7e, 0xeb, 0x7c, 0xeb, 0x7a, 0xeb, 0x78, 0xeb,
					0x76, 0xeb, 0x74, 0xeb, 0x72, 0xeb, 0x70, 0xeb, 0x6e, 0xeb, 0x6c, 0x66, 0x90, 0xeb, 0x68, 0xeb, 0x66, 0xeb, 0x64,
					0xeb, 0x62, 0xeb, 0x60, 0xeb, 0x5e, 0xeb, 0x5c, 0xeb, 0x5a, 0xeb, 0x58, 0xeb, 0x56, 0xeb, 0x54, 0xeb, 0x52, 0xeb,
					0x50, 0xeb, 0x4e, 0xeb, 0x4c, 0x66, 0x90, 0xeb, 0x48, 0xeb, 0x46, 0xeb, 0x44, 0xeb, 0x42, 0xeb, 0x40, 0xeb, 0x3e,
					0xeb, 0x3c, 0xeb, 0x3a, 0xeb, 0x38, 0xeb, 0x36, 0xeb, 0x34, 0xeb, 0x32, 0xeb, 0x30, 0xeb, 0x2e, 0xeb, 0x2c,
					0x66, 0x90, 0xeb, 0x28, 0xeb, 0x26, 0xeb, 0x24, 0xeb, 0x22, 0xeb, 0x20, 0xeb, 0x1e, 0xeb, 0x1c, 0xeb,
					0x1a, 0xeb, 0x18, 0xeb, 0x16, 0xeb, 0x14, 0xeb, 0x12, 0xeb, 0x10, 0xeb, 0xe, 0xeb, 0xc, 0x66, 0x90,
					0xeb, 0x8, 0xeb, 0x6, 0xeb, 0x4, 0xeb, 0x2, 0xeb, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JCC, exp: []byte{
					0xc3, 0xf, 0x83, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x83, 0x25, 0x2, 0x0, 0x0, 0xf, 0x83, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x83, 0x19, 0x2, 0x0, 0x0, 0xf, 0x83, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x83, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x83, 0x6, 0x2, 0x0, 0x0, 0xf, 0x83, 0x0, 0x2, 0x0, 0x0, 0xf, 0x83, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x83, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0xec, 0x1, 0x0, 0x0, 0xf, 0x83,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x83, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x83, 0xda, 0x1, 0x0, 0x0, 0xf, 0x83,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x83, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x83, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x83, 0xba, 0x1, 0x0, 0x0, 0xf, 0x83, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x83, 0xac, 0x1, 0x0, 0x0, 0xf, 0x83, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x83, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x83, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x83, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x83, 0x86, 0x1, 0x0, 0x0, 0xf, 0x83, 0x80, 0x1, 0x0, 0x0, 0xf, 0x83, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x83, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x83,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x83, 0x60, 0x1, 0x0, 0x0, 0xf, 0x83, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x83, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x83, 0x46, 0x1, 0x0, 0x0, 0xf, 0x83,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x83, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x83, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x83, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x83, 0x26, 0x1, 0x0, 0x0, 0xf, 0x83, 0x20, 0x1, 0x0, 0x0, 0xf, 0x83,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x83, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x83, 0x6, 0x1, 0x0, 0x0, 0xf, 0x83, 0x0, 0x1, 0x0, 0x0, 0xf, 0x83, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x83,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0xec, 0x0, 0x0, 0x0, 0xf, 0x83, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x83, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x83, 0xda, 0x0, 0x0, 0x0, 0xf, 0x83, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x83, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x83, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x83, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x83, 0xba, 0x0, 0x0, 0x0, 0xf, 0x83, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x83, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x83, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x83, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x83, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x83, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x83, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x83, 0x80, 0x0, 0x0, 0x0, 0x73, 0x7e, 0x73, 0x7c, 0x73, 0x7a, 0x73, 0x78, 0x73, 0x76, 0x73, 0x74,
					0x66, 0x90, 0x73, 0x70, 0x73, 0x6e, 0x73, 0x6c, 0x73, 0x6a, 0x73, 0x68, 0x73, 0x66, 0x73, 0x64, 0x73,
					0x62, 0x73, 0x60, 0x73, 0x5e, 0x73, 0x5c, 0x73, 0x5a, 0x73, 0x58, 0x73, 0x56, 0x73, 0x54, 0x66, 0x90,
					0x73, 0x50, 0x73, 0x4e, 0x73, 0x4c, 0x73, 0x4a, 0x73, 0x48, 0x73, 0x46, 0x73, 0x44, 0x73, 0x42, 0x73,
					0x40, 0x73, 0x3e, 0x73, 0x3c, 0x73, 0x3a, 0x73, 0x38, 0x73, 0x36, 0x73, 0x34, 0x66, 0x90, 0x73, 0x30,
					0x73, 0x2e, 0x73, 0x2c, 0x73, 0x2a, 0x73, 0x28, 0x73, 0x26, 0x73, 0x24, 0x73, 0x22, 0x73, 0x20, 0x73,
					0x1e, 0x73, 0x1c, 0x73, 0x1a, 0x73, 0x18, 0x73, 0x16, 0x73, 0x14, 0x66, 0x90, 0x73, 0x10, 0x73, 0xe,
					0x73, 0xc, 0x73, 0xa, 0x73, 0x8, 0x73, 0x6, 0x73, 0x4, 0x73, 0x2, 0x73, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JCS, exp: []byte{
					0xc3, 0xf, 0x82, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x82, 0x25, 0x2, 0x0, 0x0, 0xf, 0x82, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x82, 0x19, 0x2, 0x0, 0x0, 0xf, 0x82, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x82, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x82, 0x6, 0x2, 0x0, 0x0, 0xf, 0x82, 0x0, 0x2, 0x0, 0x0, 0xf, 0x82, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x82, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0xec, 0x1, 0x0, 0x0, 0xf, 0x82,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x82, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x82, 0xda, 0x1, 0x0, 0x0, 0xf, 0x82,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x82, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x82, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x82, 0xba, 0x1, 0x0, 0x0, 0xf, 0x82, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x82, 0xac, 0x1, 0x0, 0x0, 0xf, 0x82, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x82, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x82, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x82, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x82, 0x86, 0x1, 0x0, 0x0, 0xf, 0x82, 0x80, 0x1, 0x0, 0x0, 0xf, 0x82, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x82, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x82,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x82, 0x60, 0x1, 0x0, 0x0, 0xf, 0x82, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x82, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x82, 0x46, 0x1, 0x0, 0x0, 0xf, 0x82,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x82, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x82, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x82, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x82, 0x26, 0x1, 0x0, 0x0, 0xf, 0x82, 0x20, 0x1, 0x0, 0x0, 0xf, 0x82,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x82, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x82, 0x6, 0x1, 0x0, 0x0, 0xf, 0x82, 0x0, 0x1, 0x0, 0x0, 0xf, 0x82, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x82,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0xec, 0x0, 0x0, 0x0, 0xf, 0x82, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x82, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x82, 0xda, 0x0, 0x0, 0x0, 0xf, 0x82, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x82, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x82, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x82, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x82, 0xba, 0x0, 0x0, 0x0, 0xf, 0x82, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x82, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x82, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x82, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x82, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x82, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x82, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x82, 0x80, 0x0, 0x0, 0x0, 0x72, 0x7e, 0x72, 0x7c, 0x72, 0x7a, 0x72, 0x78, 0x72, 0x76, 0x72, 0x74,
					0x66, 0x90, 0x72, 0x70, 0x72, 0x6e, 0x72, 0x6c, 0x72, 0x6a, 0x72, 0x68, 0x72, 0x66, 0x72, 0x64, 0x72,
					0x62, 0x72, 0x60, 0x72, 0x5e, 0x72, 0x5c, 0x72, 0x5a, 0x72, 0x58, 0x72, 0x56, 0x72, 0x54, 0x66, 0x90,
					0x72, 0x50, 0x72, 0x4e, 0x72, 0x4c, 0x72, 0x4a, 0x72, 0x48, 0x72, 0x46, 0x72, 0x44, 0x72, 0x42, 0x72,
					0x40, 0x72, 0x3e, 0x72, 0x3c, 0x72, 0x3a, 0x72, 0x38, 0x72, 0x36, 0x72, 0x34, 0x66, 0x90, 0x72, 0x30,
					0x72, 0x2e, 0x72, 0x2c, 0x72, 0x2a, 0x72, 0x28, 0x72, 0x26, 0x72, 0x24, 0x72, 0x22, 0x72, 0x20, 0x72,
					0x1e, 0x72, 0x1c, 0x72, 0x1a, 0x72, 0x18, 0x72, 0x16, 0x72, 0x14, 0x66, 0x90, 0x72, 0x10, 0x72, 0xe,
					0x72, 0xc, 0x72, 0xa, 0x72, 0x8, 0x72, 0x6, 0x72, 0x4, 0x72, 0x2, 0x72, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JEQ, exp: []byte{
					0xc3, 0xf, 0x84, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x84, 0x25, 0x2, 0x0, 0x0, 0xf, 0x84, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x84, 0x19, 0x2, 0x0, 0x0, 0xf, 0x84, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x84, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x84, 0x6, 0x2, 0x0, 0x0, 0xf, 0x84, 0x0, 0x2, 0x0, 0x0, 0xf, 0x84, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x84, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0xec, 0x1, 0x0, 0x0, 0xf, 0x84,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x84, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x84, 0xda, 0x1, 0x0, 0x0, 0xf, 0x84,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x84, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x84, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x84, 0xba, 0x1, 0x0, 0x0, 0xf, 0x84, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x84, 0xac, 0x1, 0x0, 0x0, 0xf, 0x84, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x84, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x84, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x84, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x84, 0x86, 0x1, 0x0, 0x0, 0xf, 0x84, 0x80, 0x1, 0x0, 0x0, 0xf, 0x84, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x84, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x84,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x84, 0x60, 0x1, 0x0, 0x0, 0xf, 0x84, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x84, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x84, 0x46, 0x1, 0x0, 0x0, 0xf, 0x84,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x84, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x84, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x84, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x84, 0x26, 0x1, 0x0, 0x0, 0xf, 0x84, 0x20, 0x1, 0x0, 0x0, 0xf, 0x84,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x84, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x84, 0x6, 0x1, 0x0, 0x0, 0xf, 0x84, 0x0, 0x1, 0x0, 0x0, 0xf, 0x84, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x84,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0xec, 0x0, 0x0, 0x0, 0xf, 0x84, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x84, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x84, 0xda, 0x0, 0x0, 0x0, 0xf, 0x84, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x84, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x84, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x84, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x84, 0xba, 0x0, 0x0, 0x0, 0xf, 0x84, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x84, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x84, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x84, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x84, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x84, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x84, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x84, 0x80, 0x0, 0x0, 0x0, 0x74, 0x7e, 0x74, 0x7c, 0x74, 0x7a, 0x74, 0x78, 0x74, 0x76, 0x74, 0x74,
					0x66, 0x90, 0x74, 0x70, 0x74, 0x6e, 0x74, 0x6c, 0x74, 0x6a, 0x74, 0x68, 0x74, 0x66, 0x74, 0x64, 0x74,
					0x62, 0x74, 0x60, 0x74, 0x5e, 0x74, 0x5c, 0x74, 0x5a, 0x74, 0x58, 0x74, 0x56, 0x74, 0x54, 0x66, 0x90,
					0x74, 0x50, 0x74, 0x4e, 0x74, 0x4c, 0x74, 0x4a, 0x74, 0x48, 0x74, 0x46, 0x74, 0x44, 0x74, 0x42, 0x74,
					0x40, 0x74, 0x3e, 0x74, 0x3c, 0x74, 0x3a, 0x74, 0x38, 0x74, 0x36, 0x74, 0x34, 0x66, 0x90, 0x74, 0x30,
					0x74, 0x2e, 0x74, 0x2c, 0x74, 0x2a, 0x74, 0x28, 0x74, 0x26, 0x74, 0x24, 0x74, 0x22, 0x74, 0x20, 0x74,
					0x1e, 0x74, 0x1c, 0x74, 0x1a, 0x74, 0x18, 0x74, 0x16, 0x74, 0x14, 0x66, 0x90, 0x74, 0x10, 0x74, 0xe,
					0x74, 0xc, 0x74, 0xa, 0x74, 0x8, 0x74, 0x6, 0x74, 0x4, 0x74, 0x2, 0x74, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JGE, exp: []byte{
					0xc3, 0xf, 0x8d, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x8d, 0x25, 0x2, 0x0, 0x0, 0xf, 0x8d, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x8d, 0x19, 0x2, 0x0, 0x0, 0xf, 0x8d, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x8d, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x8d, 0x6, 0x2, 0x0, 0x0, 0xf, 0x8d, 0x0, 0x2, 0x0, 0x0, 0xf, 0x8d, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x8d, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0xec, 0x1, 0x0, 0x0, 0xf, 0x8d,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xda, 0x1, 0x0, 0x0, 0xf, 0x8d,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x8d, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xba, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x8d, 0xac, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x8d, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x86, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x80, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x8d, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x8d,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x60, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x46, 0x1, 0x0, 0x0, 0xf, 0x8d,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x8d, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x26, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x20, 0x1, 0x0, 0x0, 0xf, 0x8d,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x8d, 0x6, 0x1, 0x0, 0x0, 0xf, 0x8d, 0x0, 0x1, 0x0, 0x0, 0xf, 0x8d, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x8d,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0xec, 0x0, 0x0, 0x0, 0xf, 0x8d, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x8d, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x8d, 0xda, 0x0, 0x0, 0x0, 0xf, 0x8d, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x8d, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x8d, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x8d, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x8d, 0xba, 0x0, 0x0, 0x0, 0xf, 0x8d, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x8d, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x8d, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x8d, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x8d, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8d, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x8d, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x8d, 0x80, 0x0, 0x0, 0x0, 0x7d, 0x7e, 0x7d, 0x7c, 0x7d, 0x7a, 0x7d, 0x78, 0x7d, 0x76, 0x7d, 0x74,
					0x66, 0x90, 0x7d, 0x70, 0x7d, 0x6e, 0x7d, 0x6c, 0x7d, 0x6a, 0x7d, 0x68, 0x7d, 0x66, 0x7d, 0x64, 0x7d,
					0x62, 0x7d, 0x60, 0x7d, 0x5e, 0x7d, 0x5c, 0x7d, 0x5a, 0x7d, 0x58, 0x7d, 0x56, 0x7d, 0x54, 0x66, 0x90,
					0x7d, 0x50, 0x7d, 0x4e, 0x7d, 0x4c, 0x7d, 0x4a, 0x7d, 0x48, 0x7d, 0x46, 0x7d, 0x44, 0x7d, 0x42, 0x7d,
					0x40, 0x7d, 0x3e, 0x7d, 0x3c, 0x7d, 0x3a, 0x7d, 0x38, 0x7d, 0x36, 0x7d, 0x34, 0x66, 0x90, 0x7d, 0x30,
					0x7d, 0x2e, 0x7d, 0x2c, 0x7d, 0x2a, 0x7d, 0x28, 0x7d, 0x26, 0x7d, 0x24, 0x7d, 0x22, 0x7d, 0x20, 0x7d,
					0x1e, 0x7d, 0x1c, 0x7d, 0x1a, 0x7d, 0x18, 0x7d, 0x16, 0x7d, 0x14, 0x66, 0x90, 0x7d, 0x10, 0x7d, 0xe,
					0x7d, 0xc, 0x7d, 0xa, 0x7d, 0x8, 0x7d, 0x6, 0x7d, 0x4, 0x7d, 0x2, 0x7d, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JGT, exp: []byte{
					0xc3, 0xf, 0x8f, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x8f, 0x25, 0x2, 0x0, 0x0, 0xf, 0x8f, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x8f, 0x19, 0x2, 0x0, 0x0, 0xf, 0x8f, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x8f, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x8f, 0x6, 0x2, 0x0, 0x0, 0xf, 0x8f, 0x0, 0x2, 0x0, 0x0, 0xf, 0x8f, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x8f, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0xec, 0x1, 0x0, 0x0, 0xf, 0x8f,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xda, 0x1, 0x0, 0x0, 0xf, 0x8f,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x8f, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xba, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x8f, 0xac, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x8f, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x86, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x80, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x8f, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x8f,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x60, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x46, 0x1, 0x0, 0x0, 0xf, 0x8f,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x8f, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x26, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x20, 0x1, 0x0, 0x0, 0xf, 0x8f,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x8f, 0x6, 0x1, 0x0, 0x0, 0xf, 0x8f, 0x0, 0x1, 0x0, 0x0, 0xf, 0x8f, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x8f,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0xec, 0x0, 0x0, 0x0, 0xf, 0x8f, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x8f, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x8f, 0xda, 0x0, 0x0, 0x0, 0xf, 0x8f, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x8f, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x8f, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x8f, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x8f, 0xba, 0x0, 0x0, 0x0, 0xf, 0x8f, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x8f, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x8f, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x8f, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x8f, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8f, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x8f, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x8f, 0x80, 0x0, 0x0, 0x0, 0x7f, 0x7e, 0x7f, 0x7c, 0x7f, 0x7a, 0x7f, 0x78, 0x7f, 0x76, 0x7f, 0x74,
					0x66, 0x90, 0x7f, 0x70, 0x7f, 0x6e, 0x7f, 0x6c, 0x7f, 0x6a, 0x7f, 0x68, 0x7f, 0x66, 0x7f, 0x64, 0x7f,
					0x62, 0x7f, 0x60, 0x7f, 0x5e, 0x7f, 0x5c, 0x7f, 0x5a, 0x7f, 0x58, 0x7f, 0x56, 0x7f, 0x54, 0x66, 0x90,
					0x7f, 0x50, 0x7f, 0x4e, 0x7f, 0x4c, 0x7f, 0x4a, 0x7f, 0x48, 0x7f, 0x46, 0x7f, 0x44, 0x7f, 0x42, 0x7f,
					0x40, 0x7f, 0x3e, 0x7f, 0x3c, 0x7f, 0x3a, 0x7f, 0x38, 0x7f, 0x36, 0x7f, 0x34, 0x66, 0x90, 0x7f, 0x30,
					0x7f, 0x2e, 0x7f, 0x2c, 0x7f, 0x2a, 0x7f, 0x28, 0x7f, 0x26, 0x7f, 0x24, 0x7f, 0x22, 0x7f, 0x20, 0x7f,
					0x1e, 0x7f, 0x1c, 0x7f, 0x1a, 0x7f, 0x18, 0x7f, 0x16, 0x7f, 0x14, 0x66, 0x90, 0x7f, 0x10, 0x7f, 0xe,
					0x7f, 0xc, 0x7f, 0xa, 0x7f, 0x8, 0x7f, 0x6, 0x7f, 0x4, 0x7f, 0x2, 0x7f, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JHI, exp: []byte{
					0xc3, 0xf, 0x87, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x87, 0x25, 0x2, 0x0, 0x0, 0xf, 0x87, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x87, 0x19, 0x2, 0x0, 0x0, 0xf, 0x87, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x87, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x87, 0x6, 0x2, 0x0, 0x0, 0xf, 0x87, 0x0, 0x2, 0x0, 0x0, 0xf, 0x87, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x87, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0xec, 0x1, 0x0, 0x0, 0xf, 0x87,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x87, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x87, 0xda, 0x1, 0x0, 0x0, 0xf, 0x87,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x87, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x87, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x87, 0xba, 0x1, 0x0, 0x0, 0xf, 0x87, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x87, 0xac, 0x1, 0x0, 0x0, 0xf, 0x87, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x87, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x87, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x87, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x87, 0x86, 0x1, 0x0, 0x0, 0xf, 0x87, 0x80, 0x1, 0x0, 0x0, 0xf, 0x87, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x87, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x87,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x87, 0x60, 0x1, 0x0, 0x0, 0xf, 0x87, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x87, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x87, 0x46, 0x1, 0x0, 0x0, 0xf, 0x87,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x87, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x87, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x87, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x87, 0x26, 0x1, 0x0, 0x0, 0xf, 0x87, 0x20, 0x1, 0x0, 0x0, 0xf, 0x87,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x87, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x87, 0x6, 0x1, 0x0, 0x0, 0xf, 0x87, 0x0, 0x1, 0x0, 0x0, 0xf, 0x87, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x87,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0xec, 0x0, 0x0, 0x0, 0xf, 0x87, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x87, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x87, 0xda, 0x0, 0x0, 0x0, 0xf, 0x87, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x87, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x87, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x87, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x87, 0xba, 0x0, 0x0, 0x0, 0xf, 0x87, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x87, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x87, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x87, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x87, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x87, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x87, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x87, 0x80, 0x0, 0x0, 0x0, 0x77, 0x7e, 0x77, 0x7c, 0x77, 0x7a, 0x77, 0x78, 0x77, 0x76, 0x77, 0x74,
					0x66, 0x90, 0x77, 0x70, 0x77, 0x6e, 0x77, 0x6c, 0x77, 0x6a, 0x77, 0x68, 0x77, 0x66, 0x77, 0x64, 0x77,
					0x62, 0x77, 0x60, 0x77, 0x5e, 0x77, 0x5c, 0x77, 0x5a, 0x77, 0x58, 0x77, 0x56, 0x77, 0x54, 0x66, 0x90,
					0x77, 0x50, 0x77, 0x4e, 0x77, 0x4c, 0x77, 0x4a, 0x77, 0x48, 0x77, 0x46, 0x77, 0x44, 0x77, 0x42, 0x77,
					0x40, 0x77, 0x3e, 0x77, 0x3c, 0x77, 0x3a, 0x77, 0x38, 0x77, 0x36, 0x77, 0x34, 0x66, 0x90, 0x77, 0x30,
					0x77, 0x2e, 0x77, 0x2c, 0x77, 0x2a, 0x77, 0x28, 0x77, 0x26, 0x77, 0x24, 0x77, 0x22, 0x77, 0x20, 0x77,
					0x1e, 0x77, 0x1c, 0x77, 0x1a, 0x77, 0x18, 0x77, 0x16, 0x77, 0x14, 0x66, 0x90, 0x77, 0x10, 0x77, 0xe,
					0x77, 0xc, 0x77, 0xa, 0x77, 0x8, 0x77, 0x6, 0x77, 0x4, 0x77, 0x2, 0x77, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JLE, exp: []byte{
					0xc3, 0xf, 0x8e, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x8e, 0x25, 0x2, 0x0, 0x0, 0xf, 0x8e, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x8e, 0x19, 0x2, 0x0, 0x0, 0xf, 0x8e, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x8e, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x8e, 0x6, 0x2, 0x0, 0x0, 0xf, 0x8e, 0x0, 0x2, 0x0, 0x0, 0xf, 0x8e, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x8e, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0xec, 0x1, 0x0, 0x0, 0xf, 0x8e,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xda, 0x1, 0x0, 0x0, 0xf, 0x8e,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x8e, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xba, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x8e, 0xac, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x8e, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x86, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x80, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x8e, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x8e,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x60, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x46, 0x1, 0x0, 0x0, 0xf, 0x8e,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x8e, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x26, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x20, 0x1, 0x0, 0x0, 0xf, 0x8e,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x8e, 0x6, 0x1, 0x0, 0x0, 0xf, 0x8e, 0x0, 0x1, 0x0, 0x0, 0xf, 0x8e, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x8e,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0xec, 0x0, 0x0, 0x0, 0xf, 0x8e, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x8e, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x8e, 0xda, 0x0, 0x0, 0x0, 0xf, 0x8e, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x8e, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x8e, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x8e, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x8e, 0xba, 0x0, 0x0, 0x0, 0xf, 0x8e, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x8e, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x8e, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x8e, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x8e, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8e, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x8e, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x8e, 0x80, 0x0, 0x0, 0x0, 0x7e, 0x7e, 0x7e, 0x7c, 0x7e, 0x7a, 0x7e, 0x78, 0x7e, 0x76, 0x7e, 0x74,
					0x66, 0x90, 0x7e, 0x70, 0x7e, 0x6e, 0x7e, 0x6c, 0x7e, 0x6a, 0x7e, 0x68, 0x7e, 0x66, 0x7e, 0x64, 0x7e,
					0x62, 0x7e, 0x60, 0x7e, 0x5e, 0x7e, 0x5c, 0x7e, 0x5a, 0x7e, 0x58, 0x7e, 0x56, 0x7e, 0x54, 0x66, 0x90,
					0x7e, 0x50, 0x7e, 0x4e, 0x7e, 0x4c, 0x7e, 0x4a, 0x7e, 0x48, 0x7e, 0x46, 0x7e, 0x44, 0x7e, 0x42, 0x7e,
					0x40, 0x7e, 0x3e, 0x7e, 0x3c, 0x7e, 0x3a, 0x7e, 0x38, 0x7e, 0x36, 0x7e, 0x34, 0x66, 0x90, 0x7e, 0x30,
					0x7e, 0x2e, 0x7e, 0x2c, 0x7e, 0x2a, 0x7e, 0x28, 0x7e, 0x26, 0x7e, 0x24, 0x7e, 0x22, 0x7e, 0x20, 0x7e,
					0x1e, 0x7e, 0x1c, 0x7e, 0x1a, 0x7e, 0x18, 0x7e, 0x16, 0x7e, 0x14, 0x66, 0x90, 0x7e, 0x10, 0x7e, 0xe,
					0x7e, 0xc, 0x7e, 0xa, 0x7e, 0x8, 0x7e, 0x6, 0x7e, 0x4, 0x7e, 0x2, 0x7e, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JLS, exp: []byte{
					0xc3, 0xf, 0x86, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x86, 0x25, 0x2, 0x0, 0x0, 0xf, 0x86, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x86, 0x19, 0x2, 0x0, 0x0, 0xf, 0x86, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x86, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x86, 0x6, 0x2, 0x0, 0x0, 0xf, 0x86, 0x0, 0x2, 0x0, 0x0, 0xf, 0x86, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x86, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0xec, 0x1, 0x0, 0x0, 0xf, 0x86,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x86, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x86, 0xda, 0x1, 0x0, 0x0, 0xf, 0x86,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x86, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x86, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x86, 0xba, 0x1, 0x0, 0x0, 0xf, 0x86, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x86, 0xac, 0x1, 0x0, 0x0, 0xf, 0x86, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x86, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x86, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x86, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x86, 0x86, 0x1, 0x0, 0x0, 0xf, 0x86, 0x80, 0x1, 0x0, 0x0, 0xf, 0x86, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x86, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x86,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x86, 0x60, 0x1, 0x0, 0x0, 0xf, 0x86, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x86, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x86, 0x46, 0x1, 0x0, 0x0, 0xf, 0x86,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x86, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x86, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x86, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x86, 0x26, 0x1, 0x0, 0x0, 0xf, 0x86, 0x20, 0x1, 0x0, 0x0, 0xf, 0x86,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x86, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x86, 0x6, 0x1, 0x0, 0x0, 0xf, 0x86, 0x0, 0x1, 0x0, 0x0, 0xf, 0x86, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x86,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0xec, 0x0, 0x0, 0x0, 0xf, 0x86, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x86, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x86, 0xda, 0x0, 0x0, 0x0, 0xf, 0x86, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x86, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x86, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x86, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x86, 0xba, 0x0, 0x0, 0x0, 0xf, 0x86, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x86, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x86, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x86, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x86, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x86, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x86, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x86, 0x80, 0x0, 0x0, 0x0, 0x76, 0x7e, 0x76, 0x7c, 0x76, 0x7a, 0x76, 0x78, 0x76, 0x76, 0x76, 0x74,
					0x66, 0x90, 0x76, 0x70, 0x76, 0x6e, 0x76, 0x6c, 0x76, 0x6a, 0x76, 0x68, 0x76, 0x66, 0x76, 0x64, 0x76,
					0x62, 0x76, 0x60, 0x76, 0x5e, 0x76, 0x5c, 0x76, 0x5a, 0x76, 0x58, 0x76, 0x56, 0x76, 0x54, 0x66, 0x90,
					0x76, 0x50, 0x76, 0x4e, 0x76, 0x4c, 0x76, 0x4a, 0x76, 0x48, 0x76, 0x46, 0x76, 0x44, 0x76, 0x42, 0x76,
					0x40, 0x76, 0x3e, 0x76, 0x3c, 0x76, 0x3a, 0x76, 0x38, 0x76, 0x36, 0x76, 0x34, 0x66, 0x90, 0x76, 0x30,
					0x76, 0x2e, 0x76, 0x2c, 0x76, 0x2a, 0x76, 0x28, 0x76, 0x26, 0x76, 0x24, 0x76, 0x22, 0x76, 0x20, 0x76,
					0x1e, 0x76, 0x1c, 0x76, 0x1a, 0x76, 0x18, 0x76, 0x16, 0x76, 0x14, 0x66, 0x90, 0x76, 0x10, 0x76, 0xe,
					0x76, 0xc, 0x76, 0xa, 0x76, 0x8, 0x76, 0x6, 0x76, 0x4, 0x76, 0x2, 0x76, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JLT, exp: []byte{
					0xc3, 0xf, 0x8c, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x8c, 0x25, 0x2, 0x0, 0x0, 0xf, 0x8c, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x8c, 0x19, 0x2, 0x0, 0x0, 0xf, 0x8c, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x8c, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x8c, 0x6, 0x2, 0x0, 0x0, 0xf, 0x8c, 0x0, 0x2, 0x0, 0x0, 0xf, 0x8c, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x8c, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0xec, 0x1, 0x0, 0x0, 0xf, 0x8c,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xda, 0x1, 0x0, 0x0, 0xf, 0x8c,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x8c, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xba, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x8c, 0xac, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x8c, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x86, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x80, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x8c, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x8c,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x60, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x46, 0x1, 0x0, 0x0, 0xf, 0x8c,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x8c, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x26, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x20, 0x1, 0x0, 0x0, 0xf, 0x8c,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x8c, 0x6, 0x1, 0x0, 0x0, 0xf, 0x8c, 0x0, 0x1, 0x0, 0x0, 0xf, 0x8c, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x8c,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0xec, 0x0, 0x0, 0x0, 0xf, 0x8c, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x8c, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x8c, 0xda, 0x0, 0x0, 0x0, 0xf, 0x8c, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x8c, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x8c, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x8c, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x8c, 0xba, 0x0, 0x0, 0x0, 0xf, 0x8c, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x8c, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x8c, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x8c, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x8c, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8c, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x8c, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x8c, 0x80, 0x0, 0x0, 0x0, 0x7c, 0x7e, 0x7c, 0x7c, 0x7c, 0x7a, 0x7c, 0x78, 0x7c, 0x76, 0x7c, 0x74,
					0x66, 0x90, 0x7c, 0x70, 0x7c, 0x6e, 0x7c, 0x6c, 0x7c, 0x6a, 0x7c, 0x68, 0x7c, 0x66, 0x7c, 0x64, 0x7c,
					0x62, 0x7c, 0x60, 0x7c, 0x5e, 0x7c, 0x5c, 0x7c, 0x5a, 0x7c, 0x58, 0x7c, 0x56, 0x7c, 0x54, 0x66, 0x90,
					0x7c, 0x50, 0x7c, 0x4e, 0x7c, 0x4c, 0x7c, 0x4a, 0x7c, 0x48, 0x7c, 0x46, 0x7c, 0x44, 0x7c, 0x42, 0x7c,
					0x40, 0x7c, 0x3e, 0x7c, 0x3c, 0x7c, 0x3a, 0x7c, 0x38, 0x7c, 0x36, 0x7c, 0x34, 0x66, 0x90, 0x7c, 0x30,
					0x7c, 0x2e, 0x7c, 0x2c, 0x7c, 0x2a, 0x7c, 0x28, 0x7c, 0x26, 0x7c, 0x24, 0x7c, 0x22, 0x7c, 0x20, 0x7c,
					0x1e, 0x7c, 0x1c, 0x7c, 0x1a, 0x7c, 0x18, 0x7c, 0x16, 0x7c, 0x14, 0x66, 0x90, 0x7c, 0x10, 0x7c, 0xe,
					0x7c, 0xc, 0x7c, 0xa, 0x7c, 0x8, 0x7c, 0x6, 0x7c, 0x4, 0x7c, 0x2, 0x7c, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JMI, exp: []byte{
					0xc3, 0xf, 0x88, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x88, 0x25, 0x2, 0x0, 0x0, 0xf, 0x88, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x88, 0x19, 0x2, 0x0, 0x0, 0xf, 0x88, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x88, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x88, 0x6, 0x2, 0x0, 0x0, 0xf, 0x88, 0x0, 0x2, 0x0, 0x0, 0xf, 0x88, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x88, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0xec, 0x1, 0x0, 0x0, 0xf, 0x88,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x88, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x88, 0xda, 0x1, 0x0, 0x0, 0xf, 0x88,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x88, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x88, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x88, 0xba, 0x1, 0x0, 0x0, 0xf, 0x88, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x88, 0xac, 0x1, 0x0, 0x0, 0xf, 0x88, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x88, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x88, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x88, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x88, 0x86, 0x1, 0x0, 0x0, 0xf, 0x88, 0x80, 0x1, 0x0, 0x0, 0xf, 0x88, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x88, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x88,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x88, 0x60, 0x1, 0x0, 0x0, 0xf, 0x88, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x88, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x88, 0x46, 0x1, 0x0, 0x0, 0xf, 0x88,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x88, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x88, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x88, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x88, 0x26, 0x1, 0x0, 0x0, 0xf, 0x88, 0x20, 0x1, 0x0, 0x0, 0xf, 0x88,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x88, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x88, 0x6, 0x1, 0x0, 0x0, 0xf, 0x88, 0x0, 0x1, 0x0, 0x0, 0xf, 0x88, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x88,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0xec, 0x0, 0x0, 0x0, 0xf, 0x88, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x88, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x88, 0xda, 0x0, 0x0, 0x0, 0xf, 0x88, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x88, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x88, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x88, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x88, 0xba, 0x0, 0x0, 0x0, 0xf, 0x88, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x88, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x88, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x88, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x88, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x88, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x88, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x88, 0x80, 0x0, 0x0, 0x0, 0x78, 0x7e, 0x78, 0x7c, 0x78, 0x7a, 0x78, 0x78, 0x78, 0x76, 0x78, 0x74,
					0x66, 0x90, 0x78, 0x70, 0x78, 0x6e, 0x78, 0x6c, 0x78, 0x6a, 0x78, 0x68, 0x78, 0x66, 0x78, 0x64, 0x78,
					0x62, 0x78, 0x60, 0x78, 0x5e, 0x78, 0x5c, 0x78, 0x5a, 0x78, 0x58, 0x78, 0x56, 0x78, 0x54, 0x66, 0x90,
					0x78, 0x50, 0x78, 0x4e, 0x78, 0x4c, 0x78, 0x4a, 0x78, 0x48, 0x78, 0x46, 0x78, 0x44, 0x78, 0x42, 0x78,
					0x40, 0x78, 0x3e, 0x78, 0x3c, 0x78, 0x3a, 0x78, 0x38, 0x78, 0x36, 0x78, 0x34, 0x66, 0x90, 0x78, 0x30,
					0x78, 0x2e, 0x78, 0x2c, 0x78, 0x2a, 0x78, 0x28, 0x78, 0x26, 0x78, 0x24, 0x78, 0x22, 0x78, 0x20, 0x78,
					0x1e, 0x78, 0x1c, 0x78, 0x1a, 0x78, 0x18, 0x78, 0x16, 0x78, 0x14, 0x66, 0x90, 0x78, 0x10, 0x78, 0xe,
					0x78, 0xc, 0x78, 0xa, 0x78, 0x8, 0x78, 0x6, 0x78, 0x4, 0x78, 0x2, 0x78, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JNE, exp: []byte{
					0xc3, 0xf, 0x85, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x85, 0x25, 0x2, 0x0, 0x0, 0xf, 0x85, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x85, 0x19, 0x2, 0x0, 0x0, 0xf, 0x85, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x85, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x85, 0x6, 0x2, 0x0, 0x0, 0xf, 0x85, 0x0, 0x2, 0x0, 0x0, 0xf, 0x85, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x85, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0xec, 0x1, 0x0, 0x0, 0xf, 0x85,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x85, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x85, 0xda, 0x1, 0x0, 0x0, 0xf, 0x85,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x85, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x85, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x85, 0xba, 0x1, 0x0, 0x0, 0xf, 0x85, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x85, 0xac, 0x1, 0x0, 0x0, 0xf, 0x85, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x85, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x85, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x85, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x85, 0x86, 0x1, 0x0, 0x0, 0xf, 0x85, 0x80, 0x1, 0x0, 0x0, 0xf, 0x85, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x85, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x85,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x85, 0x60, 0x1, 0x0, 0x0, 0xf, 0x85, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x85, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x85, 0x46, 0x1, 0x0, 0x0, 0xf, 0x85,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x85, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x85, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x85, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x85, 0x26, 0x1, 0x0, 0x0, 0xf, 0x85, 0x20, 0x1, 0x0, 0x0, 0xf, 0x85,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x85, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x85, 0x6, 0x1, 0x0, 0x0, 0xf, 0x85, 0x0, 0x1, 0x0, 0x0, 0xf, 0x85, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x85,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0xec, 0x0, 0x0, 0x0, 0xf, 0x85, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x85, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x85, 0xda, 0x0, 0x0, 0x0, 0xf, 0x85, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x85, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x85, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x85, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x85, 0xba, 0x0, 0x0, 0x0, 0xf, 0x85, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x85, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x85, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x85, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x85, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x85, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x85, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x85, 0x80, 0x0, 0x0, 0x0, 0x75, 0x7e, 0x75, 0x7c, 0x75, 0x7a, 0x75, 0x78, 0x75, 0x76, 0x75, 0x74,
					0x66, 0x90, 0x75, 0x70, 0x75, 0x6e, 0x75, 0x6c, 0x75, 0x6a, 0x75, 0x68, 0x75, 0x66, 0x75, 0x64, 0x75,
					0x62, 0x75, 0x60, 0x75, 0x5e, 0x75, 0x5c, 0x75, 0x5a, 0x75, 0x58, 0x75, 0x56, 0x75, 0x54, 0x66, 0x90,
					0x75, 0x50, 0x75, 0x4e, 0x75, 0x4c, 0x75, 0x4a, 0x75, 0x48, 0x75, 0x46, 0x75, 0x44, 0x75, 0x42, 0x75,
					0x40, 0x75, 0x3e, 0x75, 0x3c, 0x75, 0x3a, 0x75, 0x38, 0x75, 0x36, 0x75, 0x34, 0x66, 0x90, 0x75, 0x30,
					0x75, 0x2e, 0x75, 0x2c, 0x75, 0x2a, 0x75, 0x28, 0x75, 0x26, 0x75, 0x24, 0x75, 0x22, 0x75, 0x20, 0x75,
					0x1e, 0x75, 0x1c, 0x75, 0x1a, 0x75, 0x18, 0x75, 0x16, 0x75, 0x14, 0x66, 0x90, 0x75, 0x10, 0x75, 0xe,
					0x75, 0xc, 0x75, 0xa, 0x75, 0x8, 0x75, 0x6, 0x75, 0x4, 0x75, 0x2, 0x75, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JPC, exp: []byte{
					0xc3, 0xf, 0x8b, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x8b, 0x25, 0x2, 0x0, 0x0, 0xf, 0x8b, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x8b, 0x19, 0x2, 0x0, 0x0, 0xf, 0x8b, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x8b, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x8b, 0x6, 0x2, 0x0, 0x0, 0xf, 0x8b, 0x0, 0x2, 0x0, 0x0, 0xf, 0x8b, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x8b, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0xec, 0x1, 0x0, 0x0, 0xf, 0x8b,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xda, 0x1, 0x0, 0x0, 0xf, 0x8b,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x8b, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xba, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x8b, 0xac, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x8b, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x86, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x80, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x8b, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x8b,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x60, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x46, 0x1, 0x0, 0x0, 0xf, 0x8b,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x8b, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x26, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x20, 0x1, 0x0, 0x0, 0xf, 0x8b,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x8b, 0x6, 0x1, 0x0, 0x0, 0xf, 0x8b, 0x0, 0x1, 0x0, 0x0, 0xf, 0x8b, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x8b,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0xec, 0x0, 0x0, 0x0, 0xf, 0x8b, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x8b, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x8b, 0xda, 0x0, 0x0, 0x0, 0xf, 0x8b, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x8b, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x8b, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x8b, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x8b, 0xba, 0x0, 0x0, 0x0, 0xf, 0x8b, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x8b, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x8b, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x8b, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x8b, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8b, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x8b, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x8b, 0x80, 0x0, 0x0, 0x0, 0x7b, 0x7e, 0x7b, 0x7c, 0x7b, 0x7a, 0x7b, 0x78, 0x7b, 0x76, 0x7b, 0x74,
					0x66, 0x90, 0x7b, 0x70, 0x7b, 0x6e, 0x7b, 0x6c, 0x7b, 0x6a, 0x7b, 0x68, 0x7b, 0x66, 0x7b, 0x64, 0x7b,
					0x62, 0x7b, 0x60, 0x7b, 0x5e, 0x7b, 0x5c, 0x7b, 0x5a, 0x7b, 0x58, 0x7b, 0x56, 0x7b, 0x54, 0x66, 0x90,
					0x7b, 0x50, 0x7b, 0x4e, 0x7b, 0x4c, 0x7b, 0x4a, 0x7b, 0x48, 0x7b, 0x46, 0x7b, 0x44, 0x7b, 0x42, 0x7b,
					0x40, 0x7b, 0x3e, 0x7b, 0x3c, 0x7b, 0x3a, 0x7b, 0x38, 0x7b, 0x36, 0x7b, 0x34, 0x66, 0x90, 0x7b, 0x30,
					0x7b, 0x2e, 0x7b, 0x2c, 0x7b, 0x2a, 0x7b, 0x28, 0x7b, 0x26, 0x7b, 0x24, 0x7b, 0x22, 0x7b, 0x20, 0x7b,
					0x1e, 0x7b, 0x1c, 0x7b, 0x1a, 0x7b, 0x18, 0x7b, 0x16, 0x7b, 0x14, 0x66, 0x90, 0x7b, 0x10, 0x7b, 0xe,
					0x7b, 0xc, 0x7b, 0xa, 0x7b, 0x8, 0x7b, 0x6, 0x7b, 0x4, 0x7b, 0x2, 0x7b, 0x0, 0xc3,
				},
			},
			{
				backward: false, jmpInst: JPS, exp: []byte{
					0xc3, 0xf, 0x8a, 0x2b, 0x2, 0x0, 0x0, 0xf, 0x8a, 0x25, 0x2, 0x0, 0x0, 0xf, 0x8a, 0x1f, 0x2, 0x0,
					0x0, 0xf, 0x8a, 0x19, 0x2, 0x0, 0x0, 0xf, 0x8a, 0x13, 0x2, 0x0, 0x0, 0x90, 0xf, 0x8a, 0xc, 0x2,
					0x0, 0x0, 0xf, 0x8a, 0x6, 0x2, 0x0, 0x0, 0xf, 0x8a, 0x0, 0x2, 0x0, 0x0, 0xf, 0x8a, 0xfa, 0x1,
					0x0, 0x0, 0xf, 0x8a, 0xf4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0xec, 0x1, 0x0, 0x0, 0xf, 0x8a,
					0xe6, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xe0, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xda, 0x1, 0x0, 0x0, 0xf, 0x8a,
					0xd4, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0xcc, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xc6, 0x1, 0x0, 0x0,
					0xf, 0x8a, 0xc0, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xba, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xb4, 0x1, 0x0, 0x0,
					0x66, 0x90, 0xf, 0x8a, 0xac, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xa6, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xa0, 0x1,
					0x0, 0x0, 0xf, 0x8a, 0x9a, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x94, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a,
					0x8c, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x86, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x80, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x7a,
					0x1, 0x0, 0x0, 0xf, 0x8a, 0x74, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0x6c, 0x1, 0x0, 0x0, 0xf, 0x8a,
					0x66, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x60, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x5a, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x54,
					0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0x4c, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x46, 0x1, 0x0, 0x0, 0xf, 0x8a,
					0x40, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x3a, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x34, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf,
					0x8a, 0x2c, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x26, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x20, 0x1, 0x0, 0x0, 0xf, 0x8a,
					0x1a, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x14, 0x1, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0xc, 0x1, 0x0, 0x0, 0xf,
					0x8a, 0x6, 0x1, 0x0, 0x0, 0xf, 0x8a, 0x0, 0x1, 0x0, 0x0, 0xf, 0x8a, 0xfa, 0x0, 0x0, 0x0, 0xf, 0x8a,
					0xf4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0xec, 0x0, 0x0, 0x0, 0xf, 0x8a, 0xe6, 0x0, 0x0, 0x0, 0xf,
					0x8a, 0xe0, 0x0, 0x0, 0x0, 0xf, 0x8a, 0xda, 0x0, 0x0, 0x0, 0xf, 0x8a, 0xd4, 0x0, 0x0, 0x0, 0x66, 0x90,
					0xf, 0x8a, 0xcc, 0x0, 0x0, 0x0, 0xf, 0x8a, 0xc6, 0x0, 0x0, 0x0, 0xf, 0x8a, 0xc0, 0x0, 0x0, 0x0, 0xf,
					0x8a, 0xba, 0x0, 0x0, 0x0, 0xf, 0x8a, 0xb4, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0xac, 0x0, 0x0, 0x0,
					0xf, 0x8a, 0xa6, 0x0, 0x0, 0x0, 0xf, 0x8a, 0xa0, 0x0, 0x0, 0x0, 0xf, 0x8a, 0x9a, 0x0, 0x0, 0x0, 0xf,
					0x8a, 0x94, 0x0, 0x0, 0x0, 0x66, 0x90, 0xf, 0x8a, 0x8c, 0x0, 0x0, 0x0, 0xf, 0x8a, 0x86, 0x0, 0x0, 0x0,
					0xf, 0x8a, 0x80, 0x0, 0x0, 0x0, 0x7a, 0x7e, 0x7a, 0x7c, 0x7a, 0x7a, 0x7a, 0x78, 0x7a, 0x76, 0x7a, 0x74,
					0x66, 0x90, 0x7a, 0x70, 0x7a, 0x6e, 0x7a, 0x6c, 0x7a, 0x6a, 0x7a, 0x68, 0x7a, 0x66, 0x7a, 0x64, 0x7a,
					0x62, 0x7a, 0x60, 0x7a, 0x5e, 0x7a, 0x5c, 0x7a, 0x5a, 0x7a, 0x58, 0x7a, 0x56, 0x7a, 0x54, 0x66, 0x90,
					0x7a, 0x50, 0x7a, 0x4e, 0x7a, 0x4c, 0x7a, 0x4a, 0x7a, 0x48, 0x7a, 0x46, 0x7a, 0x44, 0x7a, 0x42, 0x7a,
					0x40, 0x7a, 0x3e, 0x7a, 0x3c, 0x7a, 0x3a, 0x7a, 0x38, 0x7a, 0x36, 0x7a, 0x34, 0x66, 0x90, 0x7a, 0x30,
					0x7a, 0x2e, 0x7a, 0x2c, 0x7a, 0x2a, 0x7a, 0x28, 0x7a, 0x26, 0x7a, 0x24, 0x7a, 0x22, 0x7a, 0x20, 0x7a,
					0x1e, 0x7a, 0x1c, 0x7a, 0x1a, 0x7a, 0x18, 0x7a, 0x16, 0x7a, 0x14, 0x66, 0x90, 0x7a, 0x10, 0x7a, 0xe,
					0x7a, 0xc, 0x7a, 0xa, 0x7a, 0x8, 0x7a, 0x6, 0x7a, 0x4, 0x7a, 0x2, 0x7a, 0x0, 0xc3,
				},
			},
		}

		for _, tc := range tests {
			t.Run(fmt.Sprintf("%s/backward=%v", InstructionName(tc.jmpInst), tc.backward), func(t *testing.T) {
				a := NewAssembler()

				head := a.CompileStandAlone(RET)
				var jmps []asm.Node
				for i := 0; i < 128; i++ { // Large enough so that this includes long jump.
					jmps = append(jmps, a.CompileJump(tc.jmpInst))
				}
				tail := a.CompileStandAlone(RET)

				for _, jmp := range jmps {
					if tc.backward {
						jmp.AssignJumpTarget(head)
					} else {
						jmp.AssignJumpTarget(tail)
					}
				}

				actual, err := a.Assemble()
				require.NoError(t, err)
				require.Equal(t, tc.exp, actual)
			})
		}
	})
}
