package amd64

import (
	"testing"

	"github.com/tetratelabs/wazero/internal/testing/require"
)

func TestAssemblerImpl_EncodeRegisterToMemory(t *testing.T) {
	t.Run("error", func(t *testing.T) {
		tests := []struct {
			n      *nodeImpl
			expErr string
		}{
			{
				n: &nodeImpl{
					instruction: JMP,
					types:       operandTypesRegisterToMemory,
					srcReg:      RegAX, dstReg: RegAX,
				},
				expErr: "JMP is unsupported for from:register,to:memory type",
			},
			{
				n: &nodeImpl{
					instruction: SHLQ,
					types:       operandTypesRegisterToMemory,
					srcReg:      RegAX, dstReg: RegAX,
				},
				expErr: "shifting instruction SHLQ require CX register as src but got AX",
			},
		}

		for _, tc := range tests {
			a := NewAssembler()
			require.EqualError(t, a.encodeRegisterToMemory(tc.n), tc.expErr)
		}
	})

	tests := []struct {
		name string
		n    *nodeImpl
		exp  []byte
	}{
		{name: "CMPL/baseReg=R12/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: CMPL, srcReg: RegR12, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x45, 0x3b, 0x20}},
		{name: "CMPL/srcReg=R12/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: CMPL, srcReg: RegR12, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x3b, 0x24, 0x98}},
		{name: "CMPL/baseReg=R12/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: CMPL, srcReg: RegR12, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x45, 0x3b, 0x60, 0x1}},
		{name: "CMPL/srcReg=R12/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: CMPL, srcReg: RegR12, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x3b, 0x64, 0x98, 0x1}},
		{name: "CMPL/baseReg=R12/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: CMPL, srcReg: RegR12, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x45, 0x3b, 0xa0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "CMPL/srcReg=R12/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: CMPL, srcReg: RegR12, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x3b, 0xa4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "CMPL/baseReg=SI/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: CMPL, srcReg: RegSI, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x41, 0x3b, 0x30}},
		{name: "CMPL/srcReg=SI/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: CMPL, srcReg: RegSI, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x43, 0x3b, 0x34, 0x98}},
		{name: "CMPL/baseReg=SI/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: CMPL, srcReg: RegSI, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x41, 0x3b, 0x70, 0x1}},
		{name: "CMPL/srcReg=SI/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: CMPL, srcReg: RegSI, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x43, 0x3b, 0x74, 0x98, 0x1}},
		{name: "CMPL/baseReg=SI/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: CMPL, srcReg: RegSI, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x41, 0x3b, 0xb0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "CMPL/srcReg=SI/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: CMPL, srcReg: RegSI, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x43, 0x3b, 0xb4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "CMPQ/baseReg=R12/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: CMPQ, srcReg: RegR12, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x4d, 0x3b, 0x20}},
		{name: "CMPQ/srcReg=R12/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: CMPQ, srcReg: RegR12, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4f, 0x3b, 0x24, 0x98}},
		{name: "CMPQ/baseReg=R12/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: CMPQ, srcReg: RegR12, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x4d, 0x3b, 0x60, 0x1}},
		{name: "CMPQ/srcReg=R12/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: CMPQ, srcReg: RegR12, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4f, 0x3b, 0x64, 0x98, 0x1}},
		{name: "CMPQ/baseReg=R12/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: CMPQ, srcReg: RegR12, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x4d, 0x3b, 0xa0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "CMPQ/srcReg=R12/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: CMPQ, srcReg: RegR12, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4f, 0x3b, 0xa4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "CMPQ/baseReg=SI/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: CMPQ, srcReg: RegSI, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x49, 0x3b, 0x30}},
		{name: "CMPQ/srcReg=SI/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: CMPQ, srcReg: RegSI, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4b, 0x3b, 0x34, 0x98}},
		{name: "CMPQ/baseReg=SI/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: CMPQ, srcReg: RegSI, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x49, 0x3b, 0x70, 0x1}},
		{name: "CMPQ/srcReg=SI/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: CMPQ, srcReg: RegSI, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4b, 0x3b, 0x74, 0x98, 0x1}},
		{name: "CMPQ/baseReg=SI/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: CMPQ, srcReg: RegSI, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x49, 0x3b, 0xb0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "CMPQ/srcReg=SI/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: CMPQ, srcReg: RegSI, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4b, 0x3b, 0xb4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVB/baseReg=R12/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVB, srcReg: RegR12, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x45, 0x88, 0x20}},
		{name: "MOVB/srcReg=R12/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVB, srcReg: RegR12, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x88, 0x24, 0x98}},
		{name: "MOVB/baseReg=R12/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVB, srcReg: RegR12, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x45, 0x88, 0x60, 0x1}},
		{name: "MOVB/srcReg=R12/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVB, srcReg: RegR12, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x88, 0x64, 0x98, 0x1}},
		{name: "MOVB/baseReg=R12/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVB, srcReg: RegR12, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x45, 0x88, 0xa0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVB/srcReg=R12/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVB, srcReg: RegR12, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x88, 0xa4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVB/baseReg=SI/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVB, srcReg: RegSI, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x41, 0x88, 0x30}},
		{name: "MOVB/srcReg=SI/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVB, srcReg: RegSI, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x43, 0x88, 0x34, 0x98}},
		{name: "MOVB/baseReg=SI/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVB, srcReg: RegSI, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x41, 0x88, 0x70, 0x1}},
		{name: "MOVB/srcReg=SI/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVB, srcReg: RegSI, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x43, 0x88, 0x74, 0x98, 0x1}},
		{name: "MOVB/baseReg=SI/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVB, srcReg: RegSI, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x41, 0x88, 0xb0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVB/srcReg=SI/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVB, srcReg: RegSI, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x43, 0x88, 0xb4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVL/baseReg=R12/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVL, srcReg: RegR12, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x45, 0x89, 0x20}},
		{name: "MOVL/srcReg=R12/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVL, srcReg: RegR12, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x89, 0x24, 0x98}},
		{name: "MOVL/baseReg=R12/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVL, srcReg: RegR12, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x45, 0x89, 0x60, 0x1}},
		{name: "MOVL/srcReg=R12/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVL, srcReg: RegR12, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x89, 0x64, 0x98, 0x1}},
		{name: "MOVL/baseReg=R12/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVL, srcReg: RegR12, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x45, 0x89, 0xa0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVL/srcReg=R12/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVL, srcReg: RegR12, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x47, 0x89, 0xa4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVL/baseReg=X10/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVL, srcReg: RegX10, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0xf, 0x7e, 0x10}},
		{name: "MOVL/srcReg=X10/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVL, srcReg: RegX10, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0xf, 0x7e, 0x14, 0x98}},
		{name: "MOVL/baseReg=X10/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVL, srcReg: RegX10, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0xf, 0x7e, 0x50, 0x1}},
		{name: "MOVL/srcReg=X10/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVL, srcReg: RegX10, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0xf, 0x7e, 0x54, 0x98, 0x1}},
		{name: "MOVL/baseReg=X10/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVL, srcReg: RegX10, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0xf, 0x7e, 0x90, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVL/srcReg=X10/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVL, srcReg: RegX10, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0xf, 0x7e, 0x94, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVQ/baseReg=R12/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVQ, srcReg: RegR12, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x4d, 0x89, 0x20}},
		{name: "MOVQ/srcReg=R12/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVQ, srcReg: RegR12, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4f, 0x89, 0x24, 0x98}},
		{name: "MOVQ/baseReg=R12/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVQ, srcReg: RegR12, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x4d, 0x89, 0x60, 0x1}},
		{name: "MOVQ/srcReg=R12/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVQ, srcReg: RegR12, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4f, 0x89, 0x64, 0x98, 0x1}},
		{name: "MOVQ/baseReg=R12/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVQ, srcReg: RegR12, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x4d, 0x89, 0xa0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVQ/srcReg=R12/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVQ, srcReg: RegR12, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x4f, 0x89, 0xa4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVQ/baseReg=X10/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVQ, srcReg: RegX10, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0xf, 0xd6, 0x10}},
		{name: "MOVQ/srcReg=X10/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVQ, srcReg: RegX10, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0xf, 0xd6, 0x14, 0x98}},
		{name: "MOVQ/baseReg=X10/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVQ, srcReg: RegX10, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0xf, 0xd6, 0x50, 0x1}},
		{name: "MOVQ/srcReg=X10/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVQ, srcReg: RegX10, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0xf, 0xd6, 0x54, 0x98, 0x1}},
		{name: "MOVQ/baseReg=X10/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVQ, srcReg: RegX10, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0xf, 0xd6, 0x90, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVQ/srcReg=X10/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVQ, srcReg: RegX10, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0xf, 0xd6, 0x94, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVW/baseReg=R12/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVW, srcReg: RegR12, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0x89, 0x20}},
		{name: "MOVW/srcReg=R12/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVW, srcReg: RegR12, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0x89, 0x24, 0x98}},
		{name: "MOVW/baseReg=R12/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVW, srcReg: RegR12, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0x89, 0x60, 0x1}},
		{name: "MOVW/srcReg=R12/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVW, srcReg: RegR12, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0x89, 0x64, 0x98, 0x1}},
		{name: "MOVW/baseReg=R12/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVW, srcReg: RegR12, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x66, 0x45, 0x89, 0xa0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVW/srcReg=R12/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVW, srcReg: RegR12, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x47, 0x89, 0xa4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVW/baseReg=SI/offset=0x0/dstReg=R8", n: &nodeImpl{instruction: MOVW, srcReg: RegSI, dstConst: 0x0, dstReg: RegR8}, exp: []byte{0x66, 0x41, 0x89, 0x30}},
		{name: "MOVW/srcReg=SI/baseReg=R8/offset=0x0/index=R11/scale=4", n: &nodeImpl{instruction: MOVW, srcReg: RegSI, dstConst: 0x0, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x43, 0x89, 0x34, 0x98}},
		{name: "MOVW/baseReg=SI/offset=0x1/dstReg=R8", n: &nodeImpl{instruction: MOVW, srcReg: RegSI, dstConst: 0x1, dstReg: RegR8}, exp: []byte{0x66, 0x41, 0x89, 0x70, 0x1}},
		{name: "MOVW/srcReg=SI/baseReg=R8/offset=0x1/index=R11/scale=4", n: &nodeImpl{instruction: MOVW, srcReg: RegSI, dstConst: 0x1, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x43, 0x89, 0x74, 0x98, 0x1}},
		{name: "MOVW/baseReg=SI/offset=0x7fffffff/dstReg=R8", n: &nodeImpl{instruction: MOVW, srcReg: RegSI, dstConst: 0x7fffffff, dstReg: RegR8}, exp: []byte{0x66, 0x41, 0x89, 0xb0, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVW/srcReg=SI/baseReg=R8/offset=0x7fffffff/index=R11/scale=4", n: &nodeImpl{instruction: MOVW, srcReg: RegSI, dstConst: 0x7fffffff, dstMemIndex: RegR11, dstMemScale: 4, dstReg: RegR8}, exp: []byte{0x66, 0x43, 0x89, 0xb4, 0x98, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SARL/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: SARL, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x7d, 0x0}},
		{name: "SARL/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: SARL, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x7c, 0x95, 0x0}},
		{name: "SARL/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: SARL, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x7d, 0x1}},
		{name: "SARL/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: SARL, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x7c, 0x95, 0x1}},
		{name: "SARL/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: SARL, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0xbd, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SARL/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: SARL, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0xbc, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SARQ/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: SARQ, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x7d, 0x0}},
		{name: "SARQ/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: SARQ, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x7c, 0x95, 0x0}},
		{name: "SARQ/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: SARQ, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x7d, 0x1}},
		{name: "SARQ/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: SARQ, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x7c, 0x95, 0x1}},
		{name: "SARQ/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: SARQ, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0xbd, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SARQ/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: SARQ, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0xbc, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHLL/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: SHLL, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x65, 0x0}},
		{name: "SHLL/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: SHLL, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x64, 0x95, 0x0}},
		{name: "SHLL/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: SHLL, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x65, 0x1}},
		{name: "SHLL/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: SHLL, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x64, 0x95, 0x1}},
		{name: "SHLL/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: SHLL, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0xa5, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHLL/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: SHLL, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0xa4, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHLQ/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: SHLQ, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x65, 0x0}},
		{name: "SHLQ/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: SHLQ, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x64, 0x95, 0x0}},
		{name: "SHLQ/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: SHLQ, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x65, 0x1}},
		{name: "SHLQ/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: SHLQ, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x64, 0x95, 0x1}},
		{name: "SHLQ/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: SHLQ, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0xa5, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHLQ/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: SHLQ, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0xa4, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHRL/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: SHRL, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x6d, 0x0}},
		{name: "SHRL/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: SHRL, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x6c, 0x95, 0x0}},
		{name: "SHRL/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: SHRL, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x6d, 0x1}},
		{name: "SHRL/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: SHRL, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x6c, 0x95, 0x1}},
		{name: "SHRL/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: SHRL, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0xad, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHRL/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: SHRL, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0xac, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHRQ/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: SHRQ, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x6d, 0x0}},
		{name: "SHRQ/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: SHRQ, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x6c, 0x95, 0x0}},
		{name: "SHRQ/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: SHRQ, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x6d, 0x1}},
		{name: "SHRQ/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: SHRQ, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x6c, 0x95, 0x1}},
		{name: "SHRQ/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: SHRQ, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0xad, 0xff, 0xff, 0xff, 0x7f}},
		{name: "SHRQ/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: SHRQ, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0xac, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "ROLL/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: ROLL, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x45, 0x0}},
		{name: "ROLL/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: ROLL, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x44, 0x95, 0x0}},
		{name: "ROLL/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: ROLL, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x45, 0x1}},
		{name: "ROLL/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: ROLL, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x44, 0x95, 0x1}},
		{name: "ROLL/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: ROLL, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x85, 0xff, 0xff, 0xff, 0x7f}},
		{name: "ROLL/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: ROLL, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x84, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "ROLQ/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: ROLQ, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x45, 0x0}},
		{name: "ROLQ/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: ROLQ, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x44, 0x95, 0x0}},
		{name: "ROLQ/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: ROLQ, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x45, 0x1}},
		{name: "ROLQ/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: ROLQ, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x44, 0x95, 0x1}},
		{name: "ROLQ/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: ROLQ, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x85, 0xff, 0xff, 0xff, 0x7f}},
		{name: "ROLQ/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: ROLQ, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x84, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "RORL/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: RORL, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x4d, 0x0}},
		{name: "RORL/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: RORL, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x4c, 0x95, 0x0}},
		{name: "RORL/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: RORL, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x4d, 0x1}},
		{name: "RORL/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: RORL, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x4c, 0x95, 0x1}},
		{name: "RORL/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: RORL, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x41, 0xd3, 0x8d, 0xff, 0xff, 0xff, 0x7f}},
		{name: "RORL/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: RORL, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x43, 0xd3, 0x8c, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "RORQ/baseReg=CX/offset=0x0/dstReg=R13", n: &nodeImpl{instruction: RORQ, srcReg: RegCX, dstConst: 0x0, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x4d, 0x0}},
		{name: "RORQ/srcReg=CX/baseReg=R13/offset=0x0/index=R10/scale=4", n: &nodeImpl{instruction: RORQ, srcReg: RegCX, dstConst: 0x0, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x4c, 0x95, 0x0}},
		{name: "RORQ/baseReg=CX/offset=0x1/dstReg=R13", n: &nodeImpl{instruction: RORQ, srcReg: RegCX, dstConst: 0x1, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x4d, 0x1}},
		{name: "RORQ/srcReg=CX/baseReg=R13/offset=0x1/index=R10/scale=4", n: &nodeImpl{instruction: RORQ, srcReg: RegCX, dstConst: 0x1, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x4c, 0x95, 0x1}},
		{name: "RORQ/baseReg=CX/offset=0x7fffffff/dstReg=R13", n: &nodeImpl{instruction: RORQ, srcReg: RegCX, dstConst: 0x7fffffff, dstReg: RegR13}, exp: []byte{0x49, 0xd3, 0x8d, 0xff, 0xff, 0xff, 0x7f}},
		{name: "RORQ/srcReg=CX/baseReg=R13/offset=0x7fffffff/index=R10/scale=4", n: &nodeImpl{instruction: RORQ, srcReg: RegCX, dstConst: 0x7fffffff, dstMemIndex: RegR10, dstMemScale: 4, dstReg: RegR13}, exp: []byte{0x4b, 0xd3, 0x8c, 0x95, 0xff, 0xff, 0xff, 0x7f}},
		{name: "MOVDQU", n: &nodeImpl{instruction: MOVDQU, srcReg: RegX3, dstReg: RegAX, srcConst: 10}, exp: []byte{0xf3, 0xf, 0x7f, 0x18}},
		{name: "MOVDQU", n: &nodeImpl{instruction: MOVDQU, srcReg: RegX3, dstReg: RegR13, srcConst: 10}, exp: []byte{0xf3, 0x41, 0xf, 0x7f, 0x5d, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0x21, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0x21, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0x21, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0x21, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x1, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x1, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x29, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x29, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x29, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x29, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x31, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x31, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x31, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x31, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0x61, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0x61, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0x61, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0x61, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x41, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x41, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x41, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x41, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x69, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x69, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x69, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x69, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x71, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x71, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x71, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x71, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0xa1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0xa1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0xa1, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0xa1, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x81, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0x81, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x81, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0x81, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xa9, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xa9, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xa9, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xa9, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xb1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xb1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xb1, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xb1, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0xe1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x4, 0xe1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0xe1, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x44, 0xe1, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x14, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xc1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xc1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xc1, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xc1, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xe9, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xe9, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xe9, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xe9, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xf1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x4, 0xf1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xf1, 0x1, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x44, 0xf1, 0x1, 0x1}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x14, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0x21, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0x21, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0x21, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0x21, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x1, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x1, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x29, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x29, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x29, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x29, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x31, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x31, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x31, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x31, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0x61, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0x61, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0x61, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0x61, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x41, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x41, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x41, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x41, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x69, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x69, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x69, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x69, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x71, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x71, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x71, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x71, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0xa1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0xa1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0xa1, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0xa1, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x81, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0x81, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x81, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0x81, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xa9, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xa9, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xa9, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xa9, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xb1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xb1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xb1, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xb1, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0xe1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x2c, 0xe1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0xe1, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0x6c, 0xe1, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x14, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xc1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xc1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xc1, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xc1, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xe9, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xe9, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xe9, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xe9, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xf1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x2c, 0xf1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xf1, 0x1, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0x6c, 0xf1, 0x1, 0x1}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRB/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRB, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x14, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0x21, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0x21, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0x21, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0x21, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x1, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x1, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x29, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x29, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x29, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x29, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x31, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x31, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x31, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x31, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0x61, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0x61, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0x61, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0x61, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x41, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x41, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x41, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x41, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x69, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x69, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x69, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x69, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x71, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x71, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x71, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x71, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0xa1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0xa1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0xa1, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0xa1, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x81, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0x81, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x81, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0x81, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xa9, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xa9, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xa9, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xa9, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xb1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xb1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xb1, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xb1, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0xe1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x4, 0xe1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0xe1, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x44, 0xe1, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x15, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xc1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xc1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xc1, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xc1, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xe9, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xe9, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xe9, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xe9, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xf1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x4, 0xf1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xf1, 0x1, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x44, 0xf1, 0x1, 0x1}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x15, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0x21, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0x21, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0x21, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0x21, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x1, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x1, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x29, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x29, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x29, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x29, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x31, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x31, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x31, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x31, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0x61, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0x61, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0x61, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0x61, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x41, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x41, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x41, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x41, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x69, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x69, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x69, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x69, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x71, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x71, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x71, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x71, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0xa1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0xa1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0xa1, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0xa1, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x81, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0x81, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x81, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0x81, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xa9, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xa9, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xa9, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xa9, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xb1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xb1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xb1, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xb1, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0xe1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x2c, 0xe1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0xe1, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0x6c, 0xe1, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x15, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xc1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xc1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xc1, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xc1, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xe9, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xe9, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xe9, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xe9, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xf1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x2c, 0xf1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xf1, 0x1, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0x6c, 0xf1, 0x1, 0x1}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRW/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRW, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x15, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0x21, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0x21, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0x21, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0x21, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x1, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x1, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x29, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x29, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x29, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x29, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x31, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x31, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x31, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x31, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0x61, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0x61, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0x61, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0x61, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x41, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x41, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x41, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x41, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x69, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x69, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x69, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x69, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x71, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x71, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x71, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x71, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0xa1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0xa1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0xa1, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0xa1, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x81, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0x81, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x81, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0x81, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xa9, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xa9, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xa9, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xa9, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xb1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xb1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xb1, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xb1, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0xe1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x4, 0xe1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0xe1, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x44, 0xe1, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x42, 0xf, 0x3a, 0x16, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xc1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xc1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xc1, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xc1, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xe9, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xe9, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xe9, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xe9, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xf1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x4, 0xf1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xf1, 0x1, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x44, 0xf1, 0x1, 0x1}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0xf, 0x3a, 0x16, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0x21, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0x21, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0x21, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0x21, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x1, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x1, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x29, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x29, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x29, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x29, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x31, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x31, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x31, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x31, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0x61, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0x61, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0x61, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0x61, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x41, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x41, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x41, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x41, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x69, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x69, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x69, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x69, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x71, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x71, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x71, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x71, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0xa1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0xa1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0xa1, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0xa1, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x81, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0x81, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x81, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0x81, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xa9, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xa9, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xa9, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xa9, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xb1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xb1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xb1, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xb1, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0xe1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x2c, 0xe1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0xe1, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0x6c, 0xe1, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x46, 0xf, 0x3a, 0x16, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xc1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xc1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xc1, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xc1, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xe9, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xe9, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xe9, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xe9, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xf1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x2c, 0xf1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xf1, 0x1, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0x6c, 0xf1, 0x1, 0x1}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRD/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRD, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x44, 0xf, 0x3a, 0x16, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0x21, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0x21, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0x21, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0x21, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x29, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x29, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x29, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x29, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x31, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x31, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x31, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x31, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0x61, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0x61, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0x61, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0x61, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x41, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x41, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x41, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x41, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x69, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x69, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x69, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x69, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x71, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x71, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x71, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x71, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0xa1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0xa1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0xa1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0xa1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x81, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0x81, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x81, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0x81, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xa9, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xa9, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xa9, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xa9, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xb1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xb1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xb1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xb1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0xe1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x4, 0xe1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0xe1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x44, 0xe1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4a, 0xf, 0x3a, 0x16, 0x84, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xc1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xc1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xc1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xc1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xe9, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xe9, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xe9, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xe9, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xf1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x4, 0xf1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xf1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x44, 0xf1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X0/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX0, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x48, 0xf, 0x3a, 0x16, 0x84, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0x21, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0x21, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0x21, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0x21, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0x21, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x29, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x29, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x29, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x29, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x29, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x31, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=1/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x31, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x31, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=1/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x31, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=1/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 1, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x31, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0x61, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0x61, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0x61, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0x61, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0x61, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x41, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x41, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x41, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x41, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x41, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x69, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x69, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x69, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x69, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x69, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x71, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=2/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x71, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x71, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=2/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x71, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=2/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 2, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x71, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0xa1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0xa1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0xa1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0xa1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0xa1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x81, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0x81, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x81, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0x81, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0x81, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xa9, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xa9, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xa9, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xa9, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xa9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xb1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=4/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xb1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xb1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=4/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xb1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=4/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 4, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xb1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0xe1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x2c, 0xe1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0xe1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0x6c, 0xe1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=R12/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegR12, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4e, 0xf, 0x3a, 0x16, 0xac, 0xe1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xc1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xc1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xc1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xc1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=AX/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegAX, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xc1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xe9, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xe9, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xe9, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xe9, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=BP/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegBP, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xe9, 0xff, 0xff, 0xff, 0x7f, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xf1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=8/offset=0/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 0, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x2c, 0xf1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xf1, 0x1, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=8/offset=1/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 1, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0x6c, 0xf1, 0x1, 0x1}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=0", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 0}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x0}},
		{name: "PEXTRQ/src=X13/dstBase=CX/index=SI/scale=8/offset=2147483647/arg=1", n: &nodeImpl{instruction: PEXTRQ, srcReg: RegX13, dstReg: RegCX, dstMemIndex: RegSI, dstMemScale: 8, dstConst: 2147483647, arg: 1}, exp: []byte{0x66, 0x4c, 0xf, 0x3a, 0x16, 0xac, 0xf1, 0xff, 0xff, 0xff, 0x7f, 0x1}},
	}

	for _, tc := range tests {
		tc.n.types = operandTypesRegisterToMemory
		a := NewAssembler()
		err := a.encodeRegisterToMemory(tc.n)
		require.NoError(t, err, tc.name)

		actual, err := a.Assemble()
		require.NoError(t, err, tc.name)
		require.Equal(t, tc.exp, actual, tc.name)
	}
}
